LYMBO_REPORT_PATH = "LYMBO_REPORT_PATH"

LYMBO_TEST_COLLECTION = "LYMBO_TEST_COLLECTION"

LYMBO_TEST_SCOPE_CLASS = "LYMBO_TEST_SCOPE_CLASS"
LYMBO_TEST_SCOPE_FUNCTION = "LYMBO_TEST_SCOPE_FUNCTION"
LYMBO_TEST_SCOPE_MODULE = "LYMBO_TEST_SCOPE_MODULE"
LYMBO_TEST_SCOPE_GLOBAL = "LYMBO_TEST_SCOPE_GLOBAL"
LYMBO_TEST_SCOPE_MAX = "LYMBO_TEST_SCOPE_MAX"

LYMBO_LOG_LEVEL = "LYMBO_LOG_LEVEL"
LYMBO_LOG_PATH = "LYMBO_LOG_PATH"
